from dataclasses import dataclass
from typing import Any

from pychemstation.utils.tray_types import Tray


@dataclass
class Draw:
    amount: float
    source: Tray
    speed: Any
    offset: Any


@dataclass
class Wait:
    time: int


@dataclass
class Inject:
    pass


InjectorFunction = [Draw, Wait, Inject]


@dataclass
class InjectorTable:
    functions: list[InjectorFunction]
